@@ -133,6 +133,7 @@ enum
   PROP_0,
 
   PROP_N_TOUCH_POINTS,
+  PROP_THRESHOLD_TRIGGER_EDGE,
 
   PROP_LAST
 };
@@ -396,11 +397,15 @@ stage_captured_event_cb (ClutterActor       *stage,
               return CLUTTER_EVENT_PROPAGATE;
             }
 
-          if (!begin_gesture(action, actor))
+          if (!begin_gesture (action, actor))
             {
-              gesture_update_motion_point (point, event);
+              if ((point = gesture_find_point (action, event, &position)) != NULL)
+                gesture_update_motion_point (point, event);
               return CLUTTER_EVENT_PROPAGATE;
             }
+
+          if ((point = gesture_find_point (action, event, &position)) == NULL)
+            return CLUTTER_EVENT_PROPAGATE;
         }
 
       gesture_update_motion_point (point, event);
@@ -548,26 +553,6 @@ default_event_handler (ClutterGestureAction *action,
   return TRUE;
 }
 
-
-/*< private >
- * _clutter_gesture_action_set_threshold_trigger_edge:
- * @action: a #ClutterGestureAction
- * @edge: the %ClutterGestureTriggerEdge
- *
- * Sets the edge trigger for the gesture drag threshold, if any.
- *
- * This function can be called by #ClutterGestureAction subclasses that needs
- * to change the %CLUTTER_GESTURE_TRIGGER_EDGE_AFTER default.
- *
- * Since: 1.14
- */
-void
-_clutter_gesture_action_set_threshold_trigger_edge  (ClutterGestureAction     *action,
-                                                     ClutterGestureTriggerEdge edge)
-{
-  action->priv->edge = edge;
-}
-
 static void
 clutter_gesture_action_set_property (GObject      *gobject,
                                      guint         prop_id,
@@ -582,6 +567,10 @@ clutter_gesture_action_set_property (GObject      *gobject,
       clutter_gesture_action_set_n_touch_points (self, g_value_get_int (value));
       break;
 
+    case PROP_THRESHOLD_TRIGGER_EDGE:
+      clutter_gesture_action_set_threshold_trigger_edge (self, g_value_get_enum (value));
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
       break;
@@ -602,6 +591,10 @@ clutter_gesture_action_get_property (GObject    *gobject,
       g_value_set_int (value, self->priv->requested_nb_points);
       break;
 
+    case PROP_THRESHOLD_TRIGGER_EDGE:
+      g_value_set_enum (value, self->priv->edge);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
       break;
@@ -648,6 +641,24 @@ clutter_gesture_action_class_init (ClutterGestureActionClass *klass)
                       1, G_MAXINT, 1,
                       CLUTTER_PARAM_READWRITE);
 
+  /**
+   * ClutterGestureAction:threshold-trigger-edge:
+   *
+   * The trigger edge to be used by the action to either emit the
+   * #ClutterGestureAction::gesture-begin signal or to emit the
+   * #ClutterGestureAction::gesture-cancel signal.
+   *
+   * Since: 1.18
+   */
+  gesture_props[PROP_THRESHOLD_TRIGGER_EDGE] =
+    g_param_spec_enum ("threshold-trigger-edge",
+                       P_("Threshold Trigger Edge"),
+                       P_("The trigger edge used by the action"),
+                       CLUTTER_TYPE_GESTURE_TRIGGER_EDGE,
+                       CLUTTER_GESTURE_TRIGGER_EDGE_NONE,
+                       CLUTTER_PARAM_READWRITE |
+                       G_PARAM_CONSTRUCT_ONLY);
+
   g_object_class_install_properties (gobject_class,
                                      PROP_LAST,
                                      gesture_props);
@@ -754,7 +765,7 @@ clutter_gesture_action_init (ClutterGestureAction *self)
   g_array_set_clear_func (self->priv->points, (GDestroyNotify) gesture_point_unset);
 
   self->priv->requested_nb_points = 1;
-  self->priv->edge = CLUTTER_GESTURE_TRIGGER_EDGE_AFTER;
+  self->priv->edge = CLUTTER_GESTURE_TRIGGER_EDGE_NONE;
 }
 
 /**
@@ -1143,3 +1154,49 @@ clutter_gesture_action_cancel (ClutterGestureAction *action)
 
   cancel_gesture (action);
 }
+
+/**
+ * clutter_gesture_action_set_threshold_trigger_edge:
+ * @action: a #ClutterGestureAction
+ * @edge: the %ClutterGestureTriggerEdge
+ *
+ * Sets the edge trigger for the gesture drag threshold, if any.
+ *
+ * This function should only be called by sub-classes of
+ * #ClutterGestureAction during their construction phase.
+ *
+ * Since: 1.18
+ */
+void
+clutter_gesture_action_set_threshold_trigger_edge (ClutterGestureAction      *action,
+                                                   ClutterGestureTriggerEdge  edge)
+{
+  g_return_if_fail (CLUTTER_IS_GESTURE_ACTION (action));
+
+  if (action->priv->edge == edge)
+    return;
+
+  action->priv->edge = edge;
+
+  g_object_notify_by_pspec (G_OBJECT (action), gesture_props[PROP_THRESHOLD_TRIGGER_EDGE]);
+}
+
+/**
+ * clutter_gesture_action_get_threshold_trigger_egde:
+ * @action: a #ClutterGestureAction
+ *
+ * Retrieves the edge trigger of the gesture @action, as set using
+ * clutter_gesture_action_set_threshold_trigger_edge().
+ *
+ * Return value: the edge trigger
+ *
+ * Since: 1.18
+ */
+ClutterGestureTriggerEdge
+clutter_gesture_action_get_threshold_trigger_egde (ClutterGestureAction *action)
+{
+  g_return_val_if_fail (CLUTTER_IS_GESTURE_ACTION (action),
+                        CLUTTER_GESTURE_TRIGGER_EDGE_NONE);
+
+  return action->priv->edge;
+}
