@@ -39,7 +39,7 @@ G_BEGIN_DECLS
 #define CLUTTER_IS_GESTURE_ACTION(obj)            (G_TYPE_CHECK_INSTANCE_TYPE ((obj), CLUTTER_TYPE_GESTURE_ACTION))
 #define CLUTTER_GESTURE_ACTION_CLASS(klass)       (G_TYPE_CHECK_CLASS_CAST ((klass), CLUTTER_TYPE_GESTURE_ACTION, ClutterGestureActionClass))
 #define CLUTTER_IS_GESTURE_ACTION_CLASS(klass)    (G_TYPE_CHECK_CLASS_TYPE ((klass), CLUTTER_TYPE_GESTURE_ACTION))
-#define CLUTTER_GESTURE_ACTION_GET_CLASS(obj)     (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CLUTTER_TYPE_GESTURE_ACTION, ClutterGestureActionClass))
+#define CLUTTER_GESTURE_ACTION_GET_CLASS(obj)     (G_TYPE_INSTANCE_GET_CLASS ((obj), CLUTTER_TYPE_GESTURE_ACTION, ClutterGestureActionClass))
 
 typedef struct _ClutterGestureAction              ClutterGestureAction;
 typedef struct _ClutterGestureActionPrivate       ClutterGestureActionPrivate;
@@ -67,6 +67,8 @@ struct _ClutterGestureAction
  * @gesture_progress: class handler for the #ClutterGestureAction::gesture-progress signal
  * @gesture_end: class handler for the #ClutterGestureAction::gesture-end signal
  * @gesture_cancel: class handler for the #ClutterGestureAction::gesture-cancel signal
+ * @gesture_prepare: virtual function called before emitting the
+ *   #ClutterGestureAction::gesture-cancel signal
  *
  * The <structname>ClutterGestureClass</structname> structure contains only
  * private data.
@@ -87,6 +89,8 @@ struct _ClutterGestureActionClass
                                  ClutterActor          *actor);
   void     (* gesture_cancel)   (ClutterGestureAction  *action,
                                  ClutterActor          *actor);
+  gboolean (* gesture_prepare)  (ClutterGestureAction  *action,
+                                 ClutterActor          *actor);
 
   /*< private >*/
   void (* _clutter_gesture_action1) (void);
@@ -95,7 +99,6 @@ struct _ClutterGestureActionClass
   void (* _clutter_gesture_action4) (void);
   void (* _clutter_gesture_action5) (void);
   void (* _clutter_gesture_action6) (void);
-  void (* _clutter_gesture_action7) (void);
 };
 
 GType clutter_gesture_action_get_type (void) G_GNUC_CONST;
@@ -106,25 +109,25 @@ gint                   clutter_gesture_action_get_n_touch_points   (ClutterGestu
 void                   clutter_gesture_action_set_n_touch_points   (ClutterGestureAction *action,
                                                                     gint                  nb_points);
 void                   clutter_gesture_action_get_press_coords     (ClutterGestureAction *action,
-                                                                    guint                 device,
+                                                                    guint                 point,
                                                                     gfloat               *press_x,
                                                                     gfloat               *press_y);
 void                   clutter_gesture_action_get_motion_coords    (ClutterGestureAction *action,
-                                                                    guint                 device,
+                                                                    guint                 point,
                                                                     gfloat               *motion_x,
                                                                     gfloat               *motion_y);
 CLUTTER_AVAILABLE_IN_1_12
 gfloat                 clutter_gesture_action_get_motion_delta     (ClutterGestureAction *action,
-                                                                    guint                 device,
+                                                                    guint                 point,
                                                                     gfloat               *delta_x,
                                                                     gfloat               *delta_y);
 void                   clutter_gesture_action_get_release_coords   (ClutterGestureAction *action,
-                                                                    guint                 device,
+                                                                    guint                 point,
                                                                     gfloat               *release_x,
                                                                     gfloat               *release_y);
 CLUTTER_AVAILABLE_IN_1_12
 gfloat                 clutter_gesture_action_get_velocity         (ClutterGestureAction *action,
-                                                                    guint                 device,
+                                                                    guint                 point,
                                                                     gfloat               *velocity_x,
                                                                     gfloat               *velocity_y);
 
@@ -139,9 +142,19 @@ CLUTTER_AVAILABLE_IN_1_12
 ClutterInputDevice *   clutter_gesture_action_get_device           (ClutterGestureAction *action,
                                                                     guint                 point);
 
+CLUTTER_AVAILABLE_IN_1_14
+const ClutterEvent *   clutter_gesture_action_get_last_event       (ClutterGestureAction *action,
+                                                                    guint                 point);
+
 CLUTTER_AVAILABLE_IN_1_12
 void                   clutter_gesture_action_cancel               (ClutterGestureAction *action);
 
+CLUTTER_AVAILABLE_IN_1_18
+void                            clutter_gesture_action_set_threshold_trigger_edge       (ClutterGestureAction      *action,
+                                                                                         ClutterGestureTriggerEdge  edge);
+CLUTTER_AVAILABLE_IN_1_18
+ClutterGestureTriggerEdge       clutter_gesture_action_get_threshold_trigger_egde       (ClutterGestureAction      *action);
+
 G_END_DECLS
 
 #endif /* __CLUTTER_GESTURE_ACTION_H__ */
