@@ -43,6 +43,7 @@
 
 #include "clutter-debug.h"
 #include "clutter-enum-types.h"
+#include "clutter-gesture-action-private.h"
 #include "clutter-marshal.h"
 #include "clutter-private.h"
 
@@ -62,8 +63,7 @@ enum
 
 static guint rotate_signals[LAST_SIGNAL] = { 0, };
 
-G_DEFINE_TYPE (ClutterRotateAction, clutter_rotate_action,
-               CLUTTER_TYPE_GESTURE_ACTION);
+G_DEFINE_TYPE_WITH_PRIVATE (ClutterRotateAction, clutter_rotate_action, CLUTTER_TYPE_GESTURE_ACTION)
 
 static gboolean
 clutter_rotate_action_real_rotate (ClutterRotateAction *action,
@@ -178,8 +178,6 @@ clutter_rotate_action_class_init (ClutterRotateActionClass *klass)
   ClutterGestureActionClass *gesture_class =
     CLUTTER_GESTURE_ACTION_CLASS (klass);
 
-  g_type_class_add_private (klass, sizeof (ClutterRotateActionPrivate));
-
   klass->rotate = clutter_rotate_action_real_rotate;
 
   gesture_class->gesture_begin = clutter_rotate_action_gesture_begin;
@@ -217,10 +215,13 @@ clutter_rotate_action_class_init (ClutterRotateActionClass *klass)
 static void
 clutter_rotate_action_init (ClutterRotateAction *self)
 {
-  self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, CLUTTER_TYPE_ROTATE_ACTION,
-                                            ClutterRotateActionPrivate);
+  ClutterGestureAction *gesture;
 
-  clutter_gesture_action_set_n_touch_points (CLUTTER_GESTURE_ACTION (self), 2);
+  self->priv = clutter_rotate_action_get_instance_private (self);
+
+  gesture = CLUTTER_GESTURE_ACTION (self);
+  clutter_gesture_action_set_n_touch_points (gesture, 2);
+  clutter_gesture_action_set_threshold_trigger_edge (gesture, CLUTTER_GESTURE_TRIGGER_EDGE_NONE);
 }
 
 /**
