@@ -25,6 +25,7 @@
 #include "config.h"
 #endif
 
+#include <math.h>
 #include <linux/input.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -35,6 +36,7 @@
 
 #include <glib.h>
 #include <gudev/gudev.h>
+#include <libevdev/libevdev.h>
 
 #include "clutter-backend.h"
 #include "clutter-debug.h"
@@ -60,6 +62,12 @@
 
 #define AUTOREPEAT_VALUE 2
 
+/* Taken from weston/src/evdev-touchpad.c
+   In the future, we may want to make this configurable, or
+   delegate everything to loadable input drivers.
+*/
+#define DEFAULT_HYSTERESIS_MARGIN_DENOMINATOR 700.0
+
 struct _ClutterDeviceManagerEvdevPrivate
 {
   GUdevClient *udev_client;
@@ -73,12 +81,19 @@ struct _ClutterDeviceManagerEvdevPrivate
   ClutterInputDevice *core_pointer;
   ClutterInputDevice *core_keyboard;
 
+  ClutterPointerConstrainCallback constrain_callback;
+  gpointer                        constrain_data;
+  GDestroyNotify                  constrain_data_notify;
+
   ClutterStageManager *stage_manager;
   guint stage_added_handler;
   guint stage_removed_handler;
 
   GArray *keys;
   struct xkb_state *xkb;              /* XKB state object */
+  xkb_led_index_t caps_lock_led;
+  xkb_led_index_t num_lock_led;
+  xkb_led_index_t scroll_lock_led;
   uint32_t button_state;
 };
 
@@ -91,6 +106,19 @@ static gpointer                  open_callback_data;
 
 static const gchar *subsystems[] = { "input", NULL };
 
+static const char *device_type_str[] = {
+  "pointer",            /* CLUTTER_POINTER_DEVICE */
+  "keyboard",           /* CLUTTER_KEYBOARD_DEVICE */
+  "extension",          /* CLUTTER_EXTENSION_DEVICE */
+  "joystick",           /* CLUTTER_JOYSTICK_DEVICE */
+  "tablet",             /* CLUTTER_TABLET_DEVICE */
+  "touchpad",           /* CLUTTER_TOUCHPAD_DEVICE */
+  "touchscreen",        /* CLUTTER_TOUCHSCREEN_DEVICE */
+  "pen",                /* CLUTTER_PEN_DEVICE */
+  "eraser",             /* CLUTTER_ERASER_DEVICE */
+  "cursor",             /* CLUTTER_CURSOR_DEVICE */
+};
+
 /*
  * ClutterEventSource management
  *
@@ -102,8 +130,6 @@ static const gchar *subsystems[] = { "input", NULL };
  * GSource for the device manager, each device becoming a child source. Revisit
  * this once we depend on glib >= 2.28.
  */
-
-
 static const char *option_xkb_layout = "us";
 static const char *option_xkb_variant = "";
 static const char *option_xkb_options = "";
@@ -120,6 +146,16 @@ struct _ClutterEventSource
 
   ClutterInputDeviceEvdev *device;    /* back pointer to the slave evdev device */
   GPollFD event_poll_fd;              /* file descriptor of the /dev node */
+  struct libevdev *dev;
+
+  /* For mice (and emulated for touchpads) */
+  int dx, dy;
+
+  /* For touchpads */
+  gboolean touching;
+  int margin;
+  int last_x, last_y;
+  int cur_x, cur_y;
 };
 
 static gboolean
@@ -184,15 +220,40 @@ remove_key (GArray  *keys,
 }
 
 static void
-notify_key (ClutterEventSource *source,
-            guint32             time_,
-            guint32             key,
-            guint32             state)
+sync_leds (ClutterDeviceManagerEvdev *manager_evdev)
+{
+  ClutterDeviceManagerEvdevPrivate *priv = manager_evdev->priv;
+  GSList *iter;
+  int caps_lock, num_lock, scroll_lock;
+
+  caps_lock = xkb_state_led_index_is_active (priv->xkb, priv->caps_lock_led);
+  num_lock = xkb_state_led_index_is_active (priv->xkb, priv->num_lock_led);
+  scroll_lock = xkb_state_led_index_is_active (priv->xkb, priv->scroll_lock_led);
+
+  for (iter = priv->event_sources; iter; iter = iter->next)
+    {
+      ClutterEventSource *source = iter->data;
+
+      if (libevdev_has_event_type (source->dev, EV_LED))
+	libevdev_kernel_set_led_values (source->dev,
+					LED_CAPSL, caps_lock == 1 ? LIBEVDEV_LED_ON : LIBEVDEV_LED_OFF,
+					LED_NUML, num_lock == 1 ? LIBEVDEV_LED_ON : LIBEVDEV_LED_OFF,
+					LED_SCROLLL, scroll_lock == 1 ? LIBEVDEV_LED_ON : LIBEVDEV_LED_OFF,
+					-1);
+    }
+}
+
+static void
+notify_key_device (ClutterInputDevice *input_device,
+		   guint32             time_,
+		   guint32             key,
+		   guint32             state,
+		   gboolean            update_keys)
 {
-  ClutterInputDevice *input_device = (ClutterInputDevice *) source->device;
   ClutterDeviceManagerEvdev *manager_evdev;
   ClutterStage *stage;
   ClutterEvent *event = NULL;
+  enum xkb_state_component changed_state;
 
   /* We can drop the event on the floor if no stage has been
    * associated with the device yet. */
@@ -202,33 +263,46 @@ notify_key (ClutterEventSource *source,
 
   manager_evdev = CLUTTER_DEVICE_MANAGER_EVDEV (input_device->device_manager);
 
-  /* if we have keyboard state, use it to generate the event */
-  if (manager_evdev->priv->xkb)
+  event = _clutter_key_event_new_from_evdev (input_device,
+					     manager_evdev->priv->core_keyboard,
+					     stage,
+					     manager_evdev->priv->xkb,
+					     manager_evdev->priv->button_state,
+					     time_, key, state);
+
+  /* We must be careful and not pass multiple releases to xkb, otherwise it gets
+     confused and locks the modifiers */
+  if (state != AUTOREPEAT_VALUE)
     {
-      event =
-	_clutter_key_event_new_from_evdev (input_device,
-					   manager_evdev->priv->core_keyboard,
-					   stage,
-					   manager_evdev->priv->xkb,
-					   manager_evdev->priv->button_state,
-					   time_, key, state);
+      changed_state = xkb_state_update_key (manager_evdev->priv->xkb, event->key.hardware_keycode, state ? XKB_KEY_DOWN : XKB_KEY_UP);
 
-      /* We must be careful and not pass multiple releases to xkb, otherwise it gets
-	 confused and locks the modifiers */
-      if (state != AUTOREPEAT_VALUE)
+      if (update_keys)
 	{
-	  xkb_state_update_key (manager_evdev->priv->xkb, event->key.hardware_keycode, state ? XKB_KEY_DOWN : XKB_KEY_UP);
-
 	  if (state)
 	    add_key (manager_evdev->priv->keys, event->key.hardware_keycode);
 	  else
 	    remove_key (manager_evdev->priv->keys, event->key.hardware_keycode);
 	}
-
-      queue_event (event);
     }
+  else
+    changed_state = 0;
+
+  queue_event (event);
+
+  if (update_keys && (changed_state & XKB_STATE_LEDS))
+    sync_leds (manager_evdev);
 }
 
+static void
+notify_key (ClutterEventSource *source,
+            guint32             time_,
+            guint32             key,
+            guint32             state)
+{
+  ClutterInputDevice *input_device = (ClutterInputDevice *) source->device;
+
+  notify_key_device (input_device, time_, key, state, TRUE);
+}
 
 static void
 notify_relative_motion (ClutterEventSource *source,
@@ -259,14 +333,22 @@ notify_relative_motion (ClutterEventSource *source,
   clutter_input_device_get_coords (manager_evdev->priv->core_pointer, NULL, &point);
   new_x = point.x + dx;
   new_y = point.y + dy;
-  new_x = CLAMP (new_x, 0.f, stage_width - 1);
-  new_y = CLAMP (new_y, 0.f, stage_height - 1);
+
+  if (manager_evdev->priv->constrain_callback)
+    {
+      manager_evdev->priv->constrain_callback (manager_evdev->priv->core_pointer, time_, &new_x, &new_y,
+					       manager_evdev->priv->constrain_data);
+    }
+  else
+    {
+      new_x = CLAMP (new_x, 0.f, stage_width - 1);
+      new_y = CLAMP (new_y, 0.f, stage_height - 1);
+    }
 
   event->motion.time = time_;
   event->motion.stage = stage;
   event->motion.device = manager_evdev->priv->core_pointer;
-  event->motion.modifier_state = xkb_state_serialize_mods (manager_evdev->priv->xkb, XKB_STATE_EFFECTIVE);
-  event->motion.modifier_state |= manager_evdev->priv->button_state;
+  _clutter_xkb_translate_state (event, manager_evdev->priv->xkb, manager_evdev->priv->button_state);
   event->motion.x = new_x;
   event->motion.y = new_y;
   clutter_event_set_source_device (event, input_device);
@@ -275,9 +357,10 @@ notify_relative_motion (ClutterEventSource *source,
 }
 
 static void
-notify_scroll (ClutterEventSource *source,
-	       guint32             time_,
-	       gint32              value)
+notify_scroll (ClutterEventSource  *source,
+	       guint32              time_,
+	       gint32               value,
+	       ClutterOrientation   orientation)
 {
   ClutterInputDevice *input_device = (ClutterInputDevice *) source->device;
   ClutterDeviceManagerEvdev *manager_evdev;
@@ -298,9 +381,11 @@ notify_scroll (ClutterEventSource *source,
   event->scroll.time = time_;
   event->scroll.stage = CLUTTER_STAGE (stage);
   event->scroll.device = manager_evdev->priv->core_pointer;
-  event->motion.modifier_state = xkb_state_serialize_mods (manager_evdev->priv->xkb, XKB_STATE_EFFECTIVE);
-  event->scroll.modifier_state |= manager_evdev->priv->button_state;
-  event->scroll.direction = value < 0 ? CLUTTER_SCROLL_DOWN : CLUTTER_SCROLL_UP;
+  _clutter_xkb_translate_state (event, manager_evdev->priv->xkb, manager_evdev->priv->button_state);
+  if (orientation == CLUTTER_ORIENTATION_VERTICAL)
+    event->scroll.direction = value < 0 ? CLUTTER_SCROLL_DOWN : CLUTTER_SCROLL_UP;
+  else
+    event->scroll.direction = value < 0 ? CLUTTER_SCROLL_LEFT : CLUTTER_SCROLL_RIGHT;
   clutter_input_device_get_coords (manager_evdev->priv->core_pointer, NULL, &point);
   event->scroll.x = point.x;
   event->scroll.y = point.y;
@@ -377,8 +462,7 @@ notify_button (ClutterEventSource *source,
   event->button.time = time_;
   event->button.stage = CLUTTER_STAGE (stage);
   event->button.device = manager_evdev->priv->core_pointer;
-  event->motion.modifier_state = xkb_state_serialize_mods (manager_evdev->priv->xkb, XKB_STATE_EFFECTIVE);
-  event->button.modifier_state |= manager_evdev->priv->button_state;
+  _clutter_xkb_translate_state (event, manager_evdev->priv->xkb, manager_evdev->priv->button_state);
   event->button.button = button_nr;
   clutter_input_device_get_coords (manager_evdev->priv->core_pointer, NULL, &point);
   event->button.x = point.x;
@@ -388,6 +472,177 @@ notify_button (ClutterEventSource *source,
   queue_event (event);
 }
 
+static void
+process_absolute_motion (ClutterEventSource *source)
+{
+  /* Compute deltas and update absolute positions */
+  if (source->touching)
+    {
+      source->dx = source->cur_x - source->last_x;
+      source->dy = source->cur_y - source->last_y;
+      source->last_x = source->cur_x;
+      source->last_y = source->cur_y;
+    }
+}
+
+static void
+process_relative_motion (ClutterEventSource *source,
+			 guint32             _time)
+{
+  /* Flush accumulated motion deltas */
+  if (source->dx != 0 || source->dy != 0)
+    {
+      notify_relative_motion (source, _time, source->dx, source->dy);
+      source->dx = 0;
+      source->dy = 0;
+    }
+}
+
+static inline int
+hysteresis (int in, int center, int margin)
+{
+  int diff = in - center;
+
+  if (ABS (diff) <= margin)
+    return center;
+
+  if (diff > margin)
+    return center + diff - margin;
+  else if (diff < -margin)
+    return center + diff + margin;
+  return center + diff;
+}
+
+static void
+dispatch_one_event (ClutterEventSource *source,
+		    struct input_event *e)
+{
+  guint32 _time;
+
+  _time = e->time.tv_sec * 1000 + e->time.tv_usec / 1000;
+
+  switch (e->type)
+    {
+    case EV_KEY:
+      if (e->code < BTN_MISC || (e->code >= KEY_OK && e->code < BTN_TRIGGER_HAPPY))
+	notify_key (source, _time, e->code, e->value);
+      else if (e->code >= BTN_MOUSE && e->code < BTN_JOYSTICK)
+	{
+	  /* don't repeat mouse buttons */
+	  if (e->value != AUTOREPEAT_VALUE)
+	    notify_button (source, _time, e->code, e->value);
+	}
+      else if (e->code == BTN_TOOL_FINGER && e->value != AUTOREPEAT_VALUE)
+	{
+	  source->touching = e->value;
+	  if (e->value)
+	    {
+	      source->last_x = source->cur_x;
+	      source->last_y = source->cur_y;
+	    }
+	}
+      else
+	{
+	  /* We don't know about this code, ignore */
+	}
+      break;
+
+    case EV_SYN:
+      if (e->code == SYN_REPORT)
+	{
+	  process_absolute_motion (source);
+	  process_relative_motion (source, _time);
+	}
+      break;
+
+    case EV_MSC:
+      /* Nothing to do here */
+      break;
+
+    case EV_REL:
+      /* compress the EV_REL events in dx/dy */
+      switch (e->code)
+	{
+	case REL_X:
+	  source->dx += e->value;
+	  break;
+	case REL_Y:
+	  source->dy += e->value;
+	  break;
+
+	case REL_WHEEL:
+	  notify_scroll (source, _time, e->value, CLUTTER_ORIENTATION_VERTICAL);
+	  break;
+	case REL_HWHEEL:
+	  notify_scroll (source, _time, e->value, CLUTTER_ORIENTATION_HORIZONTAL);
+	  break;
+	}
+      break;
+
+    case EV_ABS:
+      switch (e->code)
+	{
+	case ABS_X:
+	  source->cur_x = hysteresis (e->value, source->cur_x, source->margin);
+	  break;
+	case ABS_Y:
+	  source->cur_y = hysteresis (e->value, source->cur_y, source->margin);
+	  break;
+	}
+      break;
+
+    default:
+      g_warning ("Unhandled event of type %d", e->type);
+      break;
+    }
+}
+
+static void
+sync_source (ClutterEventSource *source)
+{
+  struct input_event ev;
+  int err;
+  const gchar *device_path;
+
+  /* We read a SYN_DROPPED, ignore it and sync the device */
+  err = libevdev_next_event (source->dev, LIBEVDEV_READ_FLAG_SYNC, &ev);
+  while (err == 1)
+    {
+      dispatch_one_event (source, &ev);
+      err = libevdev_next_event (source->dev, LIBEVDEV_READ_FLAG_SYNC, &ev);
+    }
+
+  if (err != -EAGAIN && CLUTTER_HAS_DEBUG (EVENT))
+    {
+      device_path = _clutter_input_device_evdev_get_device_path (source->device);
+
+      CLUTTER_NOTE (EVENT, "Could not sync device (%s).", device_path);
+    }
+}
+
+static void
+fail_source (ClutterEventSource *source,
+	     int                 error)
+{
+  ClutterDeviceManager *manager;
+  ClutterInputDevice *device;
+  const gchar *device_path;
+
+  device = CLUTTER_INPUT_DEVICE (source->device);
+
+  if (CLUTTER_HAS_DEBUG (EVENT))
+    {
+      device_path = _clutter_input_device_evdev_get_device_path (source->device);
+
+      CLUTTER_NOTE (EVENT, "Could not read device (%s): %s. Removing.",
+		    device_path, strerror (error));
+    }
+
+  /* remove the faulty device */
+  manager = clutter_device_manager_get_default ();
+  _clutter_device_manager_remove_device (manager, device);
+}
+
 static gboolean
 clutter_event_dispatch (GSource     *g_source,
                         GSourceFunc  callback,
@@ -395,11 +650,10 @@ clutter_event_dispatch (GSource     *g_source,
 {
   ClutterEventSource *source = (ClutterEventSource *) g_source;
   ClutterInputDevice *input_device = (ClutterInputDevice *) source->device;
-  struct input_event ev[8];
+  struct input_event ev;
   ClutterEvent *event;
-  gint len, i, dx = 0, dy = 0;
-  uint32_t _time;
   ClutterStage *stage;
+  int err;
 
   _clutter_threads_acquire_lock ();
 
@@ -407,133 +661,50 @@ clutter_event_dispatch (GSource     *g_source,
 
   /* Don't queue more events if we haven't finished handling the previous batch
    */
-  if (!clutter_events_pending ())
+  if (clutter_events_pending ())
+    goto queue_event;
+
+  err = libevdev_next_event (source->dev, LIBEVDEV_READ_FLAG_NORMAL, &ev);
+  while (err != -EAGAIN)
     {
-       len = read (source->event_poll_fd.fd, &ev, sizeof (ev));
-       if (len < 0 || len % sizeof (ev[0]) != 0)
-       {
-         if (errno != EAGAIN)
-           {
-             ClutterDeviceManager *manager;
-             ClutterInputDevice *device;
-             const gchar *device_path;
+      if (err == 1)
+	sync_source (source);
+      else if (err == 0)
+	dispatch_one_event (source, &ev);
+      else
+	{
+	  fail_source (source, -err);
+	  goto out;
+	}
 
-             device = CLUTTER_INPUT_DEVICE (source->device);
-
-             if (CLUTTER_HAS_DEBUG (EVENT))
-               {
-                 device_path =
-                   _clutter_input_device_evdev_get_device_path (source->device);
-
-                 CLUTTER_NOTE (EVENT, "Could not read device (%s), removing.",
-                               device_path);
-               }
-
-             /* remove the faulty device */
-             manager = clutter_device_manager_get_default ();
-             _clutter_device_manager_remove_device (manager, device);
-
-           }
-         goto out;
-       }
-
-       /* Drop events if we don't have any stage to forward them to */
-       if (!stage)
-         goto out;
-
-       for (i = 0; i < len / sizeof (ev[0]); i++)
-         {
-           struct input_event *e = &ev[i];
-
-           _time = e->time.tv_sec * 1000 + e->time.tv_usec / 1000;
-
-           switch (e->type)
-             {
-             case EV_KEY:
-
-               /* don't repeat mouse buttons */
-               if (e->code >= BTN_MOUSE && e->code < KEY_OK)
-                 if (e->value == AUTOREPEAT_VALUE)
-                   continue;
-
-               switch (e->code)
-                 {
-                 case BTN_TOUCH:
-                 case BTN_TOOL_PEN:
-                 case BTN_TOOL_RUBBER:
-                 case BTN_TOOL_BRUSH:
-                 case BTN_TOOL_PENCIL:
-                 case BTN_TOOL_AIRBRUSH:
-                 case BTN_TOOL_FINGER:
-                 case BTN_TOOL_MOUSE:
-                 case BTN_TOOL_LENS:
-                   break;
-
-                 case BTN_LEFT:
-                 case BTN_RIGHT:
-                 case BTN_MIDDLE:
-                 case BTN_SIDE:
-                 case BTN_EXTRA:
-                 case BTN_FORWARD:
-                 case BTN_BACK:
-                 case BTN_TASK:
-                   notify_button(source, _time, e->code, e->value);
-                   break;
-
-                 default:
-                   notify_key (source, _time, e->code, e->value);
-                 break;
-                 }
-               break;
-
-             case EV_SYN:
-               /* Nothing to do here? */
-               break;
-
-             case EV_MSC:
-               /* Nothing to do here? */
-               break;
-
-             case EV_REL:
-               /* compress the EV_REL events in dx/dy */
-               switch (e->code)
-                 {
-                 case REL_X:
-                   dx += e->value;
-                   break;
-                 case REL_Y:
-                   dy += e->value;
-                   break;
-
-		 /* Note: we assume that REL_WHEEL is for *vertical* scroll wheels.
-		    To implement horizontal scroll, we'll need a different enum
-		    value.
-		 */
-		 case REL_WHEEL:
-		   notify_scroll (source, _time, e->value);
-		   break;
-                 }
-               break;
-
-             case EV_ABS:
-             default:
-               g_warning ("Unhandled event of type %d", e->type);
-               break;
-             }
-         }
-
-       if (dx != 0 || dy != 0)
-         notify_relative_motion (source, _time, dx, dy);
+      err = libevdev_next_event (source->dev, LIBEVDEV_READ_FLAG_NORMAL, &ev);
     }
 
+ queue_event:
+  /* Drop events if we don't have any stage to forward them to */
+  if (stage == NULL)
+    goto out;
+
   /* Pop an event off the queue if any */
   event = clutter_event_get ();
 
   if (event)
     {
+      ClutterModifierType event_state;
+      ClutterInputDevice *input_device;
+      ClutterDeviceManagerEvdev *manager_evdev;
+
+      input_device = CLUTTER_INPUT_DEVICE (source->device);
+      manager_evdev = CLUTTER_DEVICE_MANAGER_EVDEV (input_device->device_manager);
+
       /* forward the event into clutter for emission etc. */
       clutter_do_event (event);
       clutter_event_free (event);
+
+      /* update the device states *after* the event */
+      event_state = xkb_state_serialize_mods (manager_evdev->priv->xkb, XKB_STATE_MODS_EFFECTIVE);
+      _clutter_input_device_set_state (manager_evdev->priv->core_pointer, event_state);
+      _clutter_input_device_set_state (manager_evdev->priv->core_keyboard, event_state);
     }
 
 out:
@@ -556,6 +727,7 @@ clutter_event_source_new (ClutterInputDeviceEvdev *input_device)
   const gchar *node_path;
   gint fd;
   GError *error;
+  ClutterInputDeviceType device_type;
 
   /* grab the udev input device node and open it */
   node_path = _clutter_input_device_evdev_get_device_path (input_device);
@@ -565,7 +737,7 @@ clutter_event_source_new (ClutterInputDeviceEvdev *input_device)
   if (open_callback)
     {
       error = NULL;
-      fd = open_callback (node_path, O_RDONLY | O_NONBLOCK, open_callback_data, &error);
+      fd = open_callback (node_path, O_RDWR | O_NONBLOCK, open_callback_data, &error);
 
       if (fd < 0)
 	{
@@ -576,7 +748,7 @@ clutter_event_source_new (ClutterInputDeviceEvdev *input_device)
     }
   else
     {
-      fd = open (node_path, O_RDONLY | O_NONBLOCK);
+      fd = open (node_path, O_RDWR | O_NONBLOCK);
       if (fd < 0)
 	{
 	  g_warning ("Could not open device %s: %s", node_path, strerror (errno));
@@ -589,6 +761,25 @@ clutter_event_source_new (ClutterInputDeviceEvdev *input_device)
   event_source->event_poll_fd.fd = fd;
   event_source->event_poll_fd.events = G_IO_IN;
 
+  libevdev_new_from_fd (fd, &event_source->dev);
+  libevdev_set_clock_id (event_source->dev, CLOCK_MONOTONIC);
+
+  device_type = clutter_input_device_get_device_type (CLUTTER_INPUT_DEVICE (input_device));
+  if (device_type == CLUTTER_TOUCHPAD_DEVICE)
+    {
+      const struct input_absinfo *info_x, *info_y;
+      double width, height, diagonal;
+
+      info_x = libevdev_get_abs_info (event_source->dev, ABS_X);
+      info_y = libevdev_get_abs_info (event_source->dev, ABS_Y);
+
+      width = ABS (info_x->maximum - info_x->minimum);
+      height = ABS (info_y->maximum - info_y->minimum);
+      diagonal = sqrt (width * width + height * height);
+
+      event_source->margin = diagonal / DEFAULT_HYSTERESIS_MARGIN_DENOMINATOR;
+    }
+
   /* and finally configure and attach the GSource */
   g_source_set_priority (source, CLUTTER_PRIORITY_EVENTS);
   g_source_add_poll (source, &event_source->event_poll_fd);
@@ -611,6 +802,7 @@ clutter_event_source_free (ClutterEventSource *source)
   /* ignore the return value of close, it's not like we can do something
    * about it */
   close (source->event_poll_fd.fd);
+  libevdev_free (source->dev);
 
   g_source_destroy (g_source);
   g_source_unref (g_source);
@@ -637,13 +829,15 @@ find_source_by_device (ClutterDeviceManagerEvdev *manager,
 static gboolean
 is_evdev (const gchar *sysfs_path)
 {
-  GRegex *regex;
+  static GRegex *regex = NULL;
   gboolean match;
 
-  regex = g_regex_new ("/input[0-9]+/event[0-9]+$", 0, 0, NULL);
+  /* cache the regexp */
+  if (G_UNLIKELY (regex == NULL))
+    regex = g_regex_new ("/input[0-9]+/event[0-9]+$", 0, 0, NULL);
+
   match = g_regex_match (regex, sysfs_path, 0, NULL);
 
-  g_regex_unref (regex);
   return match;
 }
 
@@ -687,6 +881,8 @@ evdev_add_device (ClutterDeviceManagerEvdev *manager_evdev,
     type = CLUTTER_TOUCHPAD_DEVICE;
   else if (g_udev_device_has_property (udev_device, "ID_INPUT_TOUCHSCREEN"))
     type = CLUTTER_TOUCHSCREEN_DEVICE;
+  else
+    type = CLUTTER_EXTENSION_DEVICE;
 
   ok = sscanf (device_file, "/dev/input/event%d", &id);
   if (ok == 1)
@@ -719,8 +915,11 @@ evdev_add_device (ClutterDeviceManagerEvdev *manager_evdev,
       _clutter_input_device_add_slave (manager_evdev->priv->core_pointer, device);
     }
 
-  CLUTTER_NOTE (EVENT, "Added device %s, type %d, sysfs %s",
-                device_file, type, sysfs_path);
+  CLUTTER_NOTE (EVENT, "Added slave device '%s' (file: %s), type %s (sysfs: '%s')",
+                device_name,
+                device_file,
+                device_type_str[type],
+                sysfs_path);
 }
 
 static ClutterInputDeviceEvdev *
@@ -742,11 +941,8 @@ find_device_by_udev_device (ClutterDeviceManagerEvdev *manager_evdev,
     {
       ClutterInputDeviceEvdev *device = l->data;
 
-      if (strcmp (sysfs_path,
-                  _clutter_input_device_evdev_get_sysfs_path (device)) == 0)
-        {
-          return device;
-        }
+      if (g_strcmp0 (sysfs_path, _clutter_input_device_evdev_get_sysfs_path (device)) == 0)
+        return device;
     }
 
   return NULL;
@@ -784,6 +980,8 @@ on_uevent (GUdevClient *client,
     evdev_add_device (manager, device);
   else if (g_strcmp0 (action, "remove") == 0)
     evdev_remove_device (manager, device);
+  else
+    CLUTTER_NOTE (EVENT, "Ignored udev action '%s'", action);
 }
 
 /*
@@ -836,13 +1034,15 @@ clutter_device_manager_evdev_remove_device (ClutterDeviceManager *manager,
       source = find_source_by_device (manager_evdev, device);
       if (G_UNLIKELY (source == NULL))
 	{
-	  g_warning ("Trying to remove a device without a source installed ?!");
+	  g_critical ("Trying to remove a device without a source installed.");
 	  return;
 	}
 
       clutter_event_source_free (source);
       priv->event_sources = g_slist_remove (priv->event_sources, source);
     }
+
+  g_object_unref (device);
 }
 
 static const GSList *
@@ -926,6 +1126,9 @@ clutter_device_manager_evdev_constructed (GObject *gobject)
   ClutterDeviceManagerEvdev *manager_evdev;
   ClutterDeviceManagerEvdevPrivate *priv;
   ClutterInputDevice *device;
+  struct xkb_context *ctx;
+  struct xkb_keymap *keymap;
+  struct xkb_rule_names names;
 
   manager_evdev = CLUTTER_DEVICE_MANAGER_EVDEV (gobject);
   priv = manager_evdev->priv;
@@ -955,10 +1158,28 @@ clutter_device_manager_evdev_constructed (GObject *gobject)
   priv->core_keyboard = device;
 
   priv->keys = g_array_new (FALSE, FALSE, sizeof (guint32));
-  priv->xkb = _clutter_xkb_state_new (NULL,
-				      option_xkb_layout,
-				      option_xkb_variant,
-				      option_xkb_options);
+
+  ctx = xkb_context_new(0);
+  g_assert (ctx);
+
+  names.rules = "evdev";
+  names.model = "pc105";
+  names.layout = option_xkb_layout;
+  names.variant = option_xkb_variant;
+  names.options = option_xkb_options;
+
+  keymap = xkb_keymap_new_from_names (ctx, &names, 0);
+  xkb_context_unref(ctx);
+  if (keymap)
+    {
+      priv->xkb = xkb_state_new (keymap);
+
+      priv->caps_lock_led = xkb_keymap_led_get_index (keymap, XKB_LED_NAME_CAPS);
+      priv->num_lock_led = xkb_keymap_led_get_index (keymap, XKB_LED_NAME_NUM);
+      priv->scroll_lock_led = xkb_keymap_led_get_index (keymap, XKB_LED_NAME_SCROLL);
+
+      xkb_keymap_unref (keymap);
+    }
 
   priv->udev_client = g_udev_client_new (subsystems);
 
@@ -1029,6 +1250,9 @@ clutter_device_manager_evdev_finalize (GObject *object)
     }
   g_slist_free (priv->event_sources);
 
+  if (priv->constrain_data_notify)
+    priv->constrain_data_notify (priv->constrain_data);
+
   G_OBJECT_CLASS (clutter_device_manager_evdev_parent_class)->finalize (object);
 }
 
@@ -1168,6 +1392,8 @@ clutter_evdev_release_devices (void)
   ClutterDeviceManagerEvdev *evdev_manager;
   ClutterDeviceManagerEvdevPrivate *priv;
   GSList *l, *next;
+  uint32_t time_;
+  unsigned i;
 
   if (!manager)
     {
@@ -1189,6 +1415,13 @@ clutter_evdev_release_devices (void)
       return;
     }
 
+  /* Fake release events for all currently pressed keys */
+  time_ = g_get_monotonic_time () / 1000;
+  for (i = 0; i < priv->keys->len; i++)
+    notify_key_device (priv->core_keyboard, time_,
+		       g_array_index (priv->keys, uint32_t, i) - 8, 0, FALSE);
+  g_array_set_size (priv->keys, 0);
+
   for (l = priv->devices; l; l = next)
     {
       ClutterInputDevice *device = l->data;
@@ -1222,6 +1455,13 @@ clutter_evdev_reclaim_devices (void)
   ClutterDeviceManager *manager = clutter_device_manager_get_default ();
   ClutterDeviceManagerEvdev *evdev_manager;
   ClutterDeviceManagerEvdevPrivate *priv;
+#define LONG_BITS (sizeof(long) * 8)
+#define NLONGS(x) (((x) + LONG_BITS - 1) / LONG_BITS)
+  unsigned long key_bits[NLONGS(KEY_CNT)];
+  unsigned long source_key_bits[NLONGS(KEY_CNT)];
+  GSList *iter;
+  int i, j, rc;
+  guint32 time_;
 
   if (!manager)
     {
@@ -1244,6 +1484,37 @@ clutter_evdev_reclaim_devices (void)
 
   priv->released = FALSE;
   clutter_device_manager_evdev_probe_devices (evdev_manager);
+
+  memset (key_bits, 0, sizeof (key_bits));
+  for (iter = priv->event_sources; iter; iter = iter->next)
+    {
+      ClutterEventSource *source = iter->data;
+      ClutterInputDevice *slave = CLUTTER_INPUT_DEVICE (source->device);
+
+      if (clutter_input_device_get_device_type (slave) == CLUTTER_KEYBOARD_DEVICE)
+	{
+	  rc = ioctl (source->event_poll_fd.fd, EVIOCGBIT(EV_KEY, sizeof (source_key_bits)), source_key_bits);
+	  if (rc < 0)
+	    continue;
+
+	  for (i = 0; i < NLONGS(KEY_CNT); i++)
+	    key_bits[i] |= source_key_bits[i];
+	}
+    }
+
+  /* Fake press events for all currently pressed keys */
+  time_ = g_get_monotonic_time () / 1000;
+  for (i = 0; i < NLONGS(KEY_CNT); i++)
+    {
+      for (j = 0; j < 8; j++)
+	{
+	  if (key_bits[i] & (1 << j))
+	    notify_key_device (priv->core_keyboard, time_, i * 8 + j, 1, TRUE);
+	}
+    }
+
+#undef LONG_BITS
+#undef NLONGS
 }
 
 /**
@@ -1268,24 +1539,6 @@ clutter_evdev_set_open_callback (ClutterOpenDeviceCallback callback,
   open_callback_data = user_data;
 }
 
-/**
- * clutter_evdev_get_keyboard_state: (skip)
- * @evdev: the #ClutterDeviceManager created by the evdev backend
- *
- * Returns the xkb state tracking object for keyboard devices.
- * The object must be treated as read only, and should be used only
- * for reading out the detailed group and modifier state.
- *
- * Return value: the #xkb_state struct
- */
-struct xkb_state *
-clutter_evdev_get_keyboard_state (ClutterDeviceManager *evdev)
-{
-  g_return_val_if_fail (CLUTTER_IS_DEVICE_MANAGER_EVDEV (evdev), NULL);
-
-  return (CLUTTER_DEVICE_MANAGER_EVDEV (evdev))->priv->xkb;
-}
-
 /**
  * clutter_evdev_set_keyboard_map: (skip)
  * @evdev: the #ClutterDeviceManager created by the evdev backend
@@ -1310,6 +1563,45 @@ clutter_evdev_set_keyboard_map (ClutterDeviceManager *evdev,
   xkb_state_unref (priv->xkb);
   priv->xkb = xkb_state_new (keymap);
 
+  priv->caps_lock_led = xkb_keymap_led_get_index (keymap, XKB_LED_NAME_CAPS);
+  priv->num_lock_led = xkb_keymap_led_get_index (keymap, XKB_LED_NAME_NUM);
+  priv->scroll_lock_led = xkb_keymap_led_get_index (keymap, XKB_LED_NAME_SCROLL);
+
   for (i = 0; i < priv->keys->len; i++)
     xkb_state_update_key (priv->xkb, g_array_index (priv->keys, guint32, i), XKB_KEY_DOWN);
+
+  sync_leds (manager_evdev);
+}
+
+/**
+ * clutter_evdev_set_pointer_constrain_callback:
+ * @evdev: the #ClutterDeviceManager created by the evdev backend
+ * @callback: the callback
+ * @user_data:
+ * @user_data_notify:
+ *
+ * Sets a callback to be invoked for every pointer motion. The callback
+ * can then modify the new pointer coordinates to constrain movement within
+ * a specific region.
+ */
+void
+clutter_evdev_set_pointer_constrain_callback (ClutterDeviceManager            *evdev,
+					      ClutterPointerConstrainCallback  callback,
+					      gpointer                         user_data,
+					      GDestroyNotify                   user_data_notify)
+{
+  ClutterDeviceManagerEvdev *manager_evdev;
+  ClutterDeviceManagerEvdevPrivate *priv;
+
+  g_return_if_fail (CLUTTER_IS_DEVICE_MANAGER_EVDEV (evdev));
+
+  manager_evdev = CLUTTER_DEVICE_MANAGER_EVDEV (evdev);
+  priv = manager_evdev->priv;
+
+  if (priv->constrain_data_notify)
+    priv->constrain_data_notify (priv->constrain_data);
+
+  priv->constrain_callback = callback;
+  priv->constrain_data = user_data;
+  priv->constrain_data_notify = user_data_notify;
 }
