@@ -319,7 +319,7 @@ add_device (ClutterDeviceManagerXI2 *manager_xi2,
    */
   g_hash_table_replace (manager_xi2->devices_by_id,
                         GINT_TO_POINTER (info->deviceid),
-                        g_object_ref (device));
+                        device);
 
   if (info->use == XIMasterPointer ||
       info->use == XIMasterKeyboard)
@@ -602,7 +602,6 @@ static gdouble *
 translate_axes (ClutterInputDevice *device,
                 gdouble             x,
                 gdouble             y,
-                ClutterStageX11    *stage_x11,
                 XIValuatorState    *valuators)
 {
   guint n_axes = clutter_input_device_get_n_axes (device);
@@ -627,11 +626,11 @@ translate_axes (ClutterInputDevice *device,
       switch (axis)
         {
         case CLUTTER_INPUT_AXIS_X:
-          retval[i] = x / stage_x11->scale_factor;
+          retval[i] = x;
           break;
 
         case CLUTTER_INPUT_AXIS_Y:
-          retval[i] = y / stage_x11->scale_factor;
+          retval[i] = y;
           break;
 
         default:
@@ -643,6 +642,17 @@ translate_axes (ClutterInputDevice *device,
   return retval;
 }
 
+static void
+translate_coords (ClutterStageX11 *stage_x11,
+                  gdouble          event_x,
+                  gdouble          event_y,
+                  gfloat          *x_out,
+                  gfloat          *y_out)
+{
+  *x_out = CLAMP (event_x, 0, stage_x11->xwin_width)  / stage_x11->scale_factor;
+  *y_out = CLAMP (event_y, 0, stage_x11->xwin_height) / stage_x11->scale_factor;
+}
+
 static gdouble
 scroll_valuators_changed (ClutterInputDevice *device,
                           XIValuatorState    *valuators,
@@ -745,7 +755,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
   XGenericEventCookie *cookie;
   XIEvent *xi_event;
   XEvent *xevent;
-  int window_scale;
 
   backend_x11 = CLUTTER_BACKEND_X11 (clutter_get_default_backend ());
 
@@ -774,11 +783,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
 
   event->any.stage = stage;
 
-  if (stage_x11 != NULL)
-    window_scale = stage_x11->scale_factor;
-  else
-    window_scale = 1;
-
   switch (xi_event->evtype)
     {
     case XI_HierarchyChanged:
@@ -933,8 +937,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
             event->scroll.stage = stage;
 
             event->scroll.time = xev->time;
-            event->scroll.x = xev->event_x / window_scale;
-            event->scroll.y = xev->event_y / window_scale;
+            translate_coords (stage_x11, xev->event_x, xev->event_y, &event->scroll.x, &event->scroll.y);
 	    _clutter_input_device_xi2_translate_state (event,
 						       &xev->mods,
 						       &xev->buttons,
@@ -946,7 +949,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
             event->scroll.axes = translate_axes (event->scroll.device,
                                                  event->scroll.x,
                                                  event->scroll.y,
-                                                 stage_x11,
                                                  &xev->valuators);
 
             CLUTTER_NOTE (EVENT,
@@ -981,8 +983,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
             event->button.stage = stage;
 
             event->button.time = xev->time;
-            event->button.x = xev->event_x / window_scale;
-            event->button.y = xev->event_y / window_scale;
+            translate_coords (stage_x11, xev->event_x, xev->event_y, &event->button.x, &event->button.y);
             event->button.button = xev->detail;
 	    _clutter_input_device_xi2_translate_state (event,
 						       &xev->mods,
@@ -995,7 +996,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
             event->button.axes = translate_axes (event->button.device,
                                                  event->button.x,
                                                  event->button.y,
-                                                 stage_x11,
                                                  &xev->valuators);
 
             CLUTTER_NOTE (EVENT,
@@ -1064,8 +1064,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
 
             event->scroll.stage = stage;
             event->scroll.time = xev->time;
-            event->scroll.x = xev->event_x / window_scale;
-            event->scroll.y = xev->event_y / window_scale;
+            translate_coords (stage_x11, xev->event_x, xev->event_y, &event->scroll.x, &event->scroll.y);
 	    _clutter_input_device_xi2_translate_state (event,
 						       &xev->mods,
 						       &xev->buttons,
@@ -1093,8 +1092,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         event->motion.stage = stage;
 
         event->motion.time = xev->time;
-        event->motion.x = xev->event_x / window_scale;
-        event->motion.y = xev->event_y / window_scale;
+        translate_coords (stage_x11, xev->event_x, xev->event_y, &event->motion.x, &event->motion.y);
 	_clutter_input_device_xi2_translate_state (event,
 						   &xev->mods,
 						   &xev->buttons,
@@ -1106,7 +1104,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         event->motion.axes = translate_axes (event->motion.device,
                                              event->motion.x,
                                              event->motion.y,
-                                             stage_x11,
                                              &xev->valuators);
 
         if (source_device != NULL && device->stage != NULL)
@@ -1145,8 +1142,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
 
         event->touch.stage = stage;
         event->touch.time = xev->time;
-        event->touch.x = xev->event_x / window_scale;
-        event->touch.y = xev->event_y / window_scale;
+        translate_coords (stage_x11, xev->event_x, xev->event_y, &event->touch.x, &event->touch.y);
 	_clutter_input_device_xi2_translate_state (event,
 						   &xev->mods,
 						   &xev->buttons,
@@ -1161,7 +1157,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         event->touch.axes = translate_axes (event->touch.device,
                                             event->motion.x,
                                             event->motion.y,
-                                            stage_x11,
                                             &xev->valuators);
 
         if (xi_event->evtype == XI_TouchBegin)
@@ -1201,8 +1196,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         event->touch.stage = stage;
         event->touch.time = xev->time;
         event->touch.sequence = GUINT_TO_POINTER (xev->detail);
-        event->touch.x = xev->event_x / window_scale;
-        event->touch.y = xev->event_y / window_scale;
+        translate_coords (stage_x11, xev->event_x, xev->event_y, &event->touch.x, &event->touch.y);
 
         clutter_event_set_source_device (event, source_device);
 
@@ -1213,7 +1207,6 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         event->touch.axes = translate_axes (event->touch.device,
                                             event->motion.x,
                                             event->motion.y,
-                                            stage_x11,
                                             &xev->valuators);
 
 	_clutter_input_device_xi2_translate_state (event,
@@ -1259,8 +1252,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
             event->crossing.related = NULL;
 
             event->crossing.time = xev->time;
-            event->crossing.x = xev->event_x / window_scale;
-            event->crossing.y = xev->event_y / window_scale;
+            translate_coords (stage_x11, xev->event_x, xev->event_y, &event->crossing.x, &event->crossing.y);
 
             _clutter_input_device_set_stage (device, stage);
           }
@@ -1283,8 +1275,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
             event->crossing.related = NULL;
 
             event->crossing.time = xev->time;
-            event->crossing.x = xev->event_x / window_scale;
-            event->crossing.y = xev->event_y / window_scale;
+            translate_coords (stage_x11, xev->event_x, xev->event_y, &event->crossing.x, &event->crossing.y);
 
             _clutter_input_device_set_stage (device, NULL);
           }
